/*
 * DetectData.java
 *
 * Copyright (C) 1998-2000 Seiichi Yoshida
 *
 */

package misao.pixy;

import java.awt.Color;
import misao.util.*;

public class DetectData extends StarData {
    private double value;
    private double peak;
    private int area;
    private double radius;
    private boolean edge_flag;
    private PSF psf;
    private boolean cosmic_ray;
    private String mag_code;
    private double check_mag;

    private static final int DATA_TYPE_DETAIL = 0;
    private static final int DATA_TYPE_SIMPLE = 1;
    private static final int DATA_TYPE_ONLY_XY = 2;
    private static final int DATA_TYPE_ONLY_XY_MAG = 3;
    private int type;

    public DetectData ( ) {
	super();

	value = 0.0;
	peak = 0.0;
	area = 0;
	radius = 0.0;
	edge_flag = false;
	psf = null;
	cosmic_ray = false;
	mag_code = "";
	check_mag = Astro.ILLEGAL_MAG;

	type = DATA_TYPE_DETAIL;
    }

    public DetectData ( DetectData sd ) {
	super(sd);

	value = sd.value;
	peak = sd.peak;
	area = sd.area;
	radius = sd.radius;
	edge_flag = sd.edge_flag;
	psf = new PSF(sd.psf);
	cosmic_ray = sd.cosmic_ray;
	mag_code = sd.mag_code;
	check_mag = sd.check_mag;

	type = DATA_TYPE_DETAIL;
    }

    public DetectData ( Position p ) {
	this();

	setPosition(p);

	type = DATA_TYPE_ONLY_XY;
    }

    public DetectData ( AdjoiningPixel ap, int pixels_for_astrometry ) {
	this();

	Xyz xyz = new Xyz();
	double peak_value = 0.0;
	for (int i = 0 ; i < ap.size() ; i++) {
	    double x = ((double)ap.elementPointAt(i).x + 0.5) * ap.elementValueAt(i);
	    double y = ((double)ap.elementPointAt(i).y + 0.5) * ap.elementValueAt(i);
	    xyz.add(new Xyz(x, y, ap.elementValueAt(i)));

	    if (peak_value < ap.elementValueAt(i))
		peak_value = ap.elementValueAt(i);
	}

	setValue(xyz.getZ());
	setPeak(peak_value);
	setArea(ap.size());
	setRadius(Math.sqrt((double)getArea() / Math.PI));
	setEdge(ap.isEdge());

	// astrometry with several top pixels
	if (ap.size() < pixels_for_astrometry) {
	    setX(xyz.getX() / xyz.getZ());
	    setY(xyz.getY() / xyz.getZ());
	} else {
	    Array array = new Array(ap.size());
	    for (int i = 0 ; i < ap.size() ; i++)
		array.set(i, ap.elementValueAt(i));
	    ArrayIndex index = array.sortDescendant();
	    xyz = new Xyz();
	    double value = 0.0;
	    for (int i = 0 ; i < pixels_for_astrometry ; i++) {
		double x = ((double)ap.elementPointAt(index.get(i)).x + 0.5) * ap.elementValueAt(index.get(i));
		double y = ((double)ap.elementPointAt(index.get(i)).y + 0.5) * ap.elementValueAt(index.get(i));
		xyz.add(new Xyz(x, y, ap.elementValueAt(index.get(i))));
		value = array.getValue(i);
	    }
	    for (int i = pixels_for_astrometry ; i < ap.size() ; i++) {
		if (array.getValue(i) > value * 0.8) {
		    double x = ((double)ap.elementPointAt(index.get(i)).x + 0.5) * ap.elementValueAt(index.get(i));
		    double y = ((double)ap.elementPointAt(index.get(i)).y + 0.5) * ap.elementValueAt(index.get(i));
		    xyz.add(new Xyz(x, y, ap.elementValueAt(index.get(i))));
		} else
		    break;
	    }
	    setX(xyz.getX() / xyz.getZ());
	    setY(xyz.getY() / xyz.getZ());
	}

	// PSF radius
	psf = new PSF(getPosition(), getPeak());
	for (int i = 0 ; i < ap.size() ; i++)
	    psf.add(new Position((double)ap.elementPointAt(i).x + 0.5, (double)ap.elementPointAt(i).y + 0.5), ap.elementValueAt(i));
	psf.solve();
    }

    public double getValue ( ) {
	return value;
    }

    public double getPeak ( ) {
	return peak;
    }

    public int getArea ( ) {
	return area;
    }

    public double getRadius ( ) {
	return radius;
    }

    public String getRecordSimple ( ) {
	type = DATA_TYPE_SIMPLE;
	return getRecord();
    }
    public String getRecordDetail ( ) {
	type = DATA_TYPE_DETAIL;
	return getRecord();
    }
    public String getRecordOnlyXYMag ( ) {
	type = DATA_TYPE_ONLY_XY_MAG;
	return getRecord();
    }
    public String getRecord ( ) {
	String record = "";
	if (type != DATA_TYPE_ONLY_XY  &&  type != DATA_TYPE_ONLY_XY_MAG) {
	    record += getCoor().toStringRA() + " " + getCoor().toStringDecl();
	    record += "  ";
	    record += "mag=" + Astro.toString(getMag(), 5, 2) + mag_code;
	    record += "  ";

	    if (check_mag < Astro.ILLEGAL_MAG - 1.0) {
		record += "Cm=" + Astro.toString(getCheckMag(), 5, 2) + mag_code;
		record += "  ";
	    }
	}
	record += "x=" + Astro.toString(getX(), 7, 4);
	record += "  ";
	record += "y=" + Astro.toString(getY(), 7, 4);
	if (type == DATA_TYPE_ONLY_XY_MAG) {
	    record += "  ";
	    record += "mag=" + Astro.toString(getMag(), 5, 2) + mag_code;

	    if (check_mag < Astro.ILLEGAL_MAG - 1.0) {
		record += "  Cm=" + Astro.toString(getCheckMag(), 5, 2) + mag_code;
	    }
	} else if (type == DATA_TYPE_DETAIL) {
	    record += "  ";
	    record += "value=" + Astro.toString(value, 10, 7);
	    record += "  ";
	    record += "peak=" + Astro.toString(peak, 9, 6);
	    record += "  ";
	    record += "area=" + Astro.toString(area, 4);
//	    record += "  ";
//	    record += "psf=" + Astro.toString(psf.getRadius(), 4, 2);
	}

	return record;
    }

    public void setValue ( double Value ) {
	value = Value;
    }

    public void setPeak ( double Peak ) {
	peak = Peak;
    }

    public void setArea ( int Area ) {
	area = Area;
    }

    public void setRadius ( double Radius ) {
	radius = Radius;
    }

    public boolean isEdge ( ) {
	return edge_flag;
    }

    public boolean isCosmicRay ( ) {
	return cosmic_ray;
    }

    public void setEdge ( boolean Edge ) {
	edge_flag = Edge;
    }

    public void markCosmicRay ( ) {
	cosmic_ray = true;
    }

    public PSF getPSF ( ) {
	return psf;
    }

    public void setMag ( double _mag, String _mag_code ) {
	setMag(_mag);
	mag_code = _mag_code;
    }

    public String getMagCode ( ) {
	return mag_code;
    }

    public double getCheckMag ( ) {
	return check_mag;
    }

    public void setCheckMag ( double _check_mag ) {
	check_mag = _check_mag;
    }

    public void setPrintDetail ( ) {
	setPrint(DATA_TYPE_DETAIL);
    }
    public void setPrintSimple ( ) {
	setPrint(DATA_TYPE_SIMPLE);
    }
    public void setPrintOnlyXYMag ( ) {
	setPrint(DATA_TYPE_ONLY_XY_MAG);
    }
    private void setPrint ( int t ) {
	type = t;
	DetectData sd = (DetectData)getFirstAppendix();
	while (sd != null) {
	    sd.type = t;
	    sd = (DetectData)getNextAppendix();
	}
    }

    public static DetectData createDetectData ( String s ) {
	DetectData data = new DetectData();

	int p = 0;
	if (s.length() >= 2  &&  s.substring(0,2).equals("x=")) {
	    data.type = DATA_TYPE_ONLY_XY;
	} else {
	    data.type = DATA_TYPE_SIMPLE;

	    p = s.indexOf("  ");
	    data.setCoor(Coor.createCoor(s.substring(0,p)));
	    s = s.substring(p+2).trim();
	}

	double x = 0.0;
	double y = 0.0;
	while (s.length() > 0) {
	    if (s.length() >= 4  &&  s.substring(0,4).equals("mag=")) {
		s = s.substring(4).trim();
		String m = s;
		p = m.indexOf(' ');
		if (p >= 0)
		    m = m.substring(0, p);
		p = m.length() - 1;
		while (p >= 0) {
		    if ('0' <= m.charAt(p)  &&  m.charAt(p) <= '9')
			break;
		    p--;
		}
		p++;
		if (p < m.length())
		    data.setMag(Astro.doubleValueOf(m.substring(0,p)), m.substring(p));
		else
		    data.setMag(Astro.doubleValueOf(s));
	    } else if (s.length() >= 2  &&  s.substring(0,2).equals("x=")) {
		s = s.substring(2).trim();
		x = Astro.doubleValueOf(s);
	    } else if (s.length() >= 2  &&  s.substring(0,2).equals("y=")) {
		s = s.substring(2).trim();
		y = Astro.doubleValueOf(s);
	    } else if (s.length() >= 6  &&  s.substring(0,6).equals("value=")) {
		data.type = DATA_TYPE_DETAIL;
		s = s.substring(6).trim();
		data.setValue(Astro.doubleValueOf(s));
	    } else if (s.length() >= 5  &&  s.substring(0,5).equals("peak=")) {
		data.type = DATA_TYPE_DETAIL;
		s = s.substring(5).trim();
		data.setPeak(Astro.doubleValueOf(s));
	    } else if (s.length() >= 5  &&  s.substring(0,5).equals("area=")) {
		data.type = DATA_TYPE_DETAIL;
		s = s.substring(5).trim();
		data.setArea((int)(Astro.doubleValueOf(s) + 0.5));
	    }

	    p = s.indexOf(' ');
	    if (p < 0)
		s = "";
	    else
		s = s.substring(p).trim();
	}
	data.setPosition(new Position(x,y));

	return data;
    }
}
