/*
 * Pixy.java
 *
 * Copyright (C) 1998-2000 Seiichi Yoshida
 *
 */

package misao.pixy;

import java.io.*;
import java.net.*;
import java.util.*;
import misao.util.*;
import misao.pixy.test.*;
import misao.host.*;
import misao.database.*;
import misao.diff.*;

public class Pixy {
    // start on console
    public static void main ( String[] args ) throws Exception {
	Version.printCopyright(System.out, "PIXY system");

	/* ------------------------------------------------------
	    java misao.pixy.Pixy
		Open an setup window.
	    java misao.pixy.Pixy -arg <arguments>
		Run PIXY with parameters in arguments.
	    java misao.pixy.Pixy -batch <batch file>
		Run PIXY with parameters in batch file.
	    java misao.pixy.Pixy -console
		Run PIXY setting parameters in Q&A style.
//	    java misao.pixy.Pixy -diff <pxf file1> <pxf file2>
	    java misao.pixy.Pixy -diff <image1> <image2> [<output file>]
		Output difference between two images.
	    java misao.pixy.Pixy -identify <pxf file> [<check file>]
		Identification with known variable stars.
	    java misao.pixy.Pixy -report <pxf file> [<report file>]
		Output a photometric report.
	    java misao.pixy.Pixy -review <pxf file> [<image>]
		Review PIXY with a PXF file.
	    java misao.pixy.Pixy -survey <directory> [<output file>]
	        List up candidates of new objects.
	    java misao.pixy.Pixy -update [-remove] <pxf file> [<image ID>]
	        Update magnitude database.

	    java misao.pixy.Pixy -test detect
		Run PIXY to check raw detected stars.
	    java misao.pixy.Pixy -test matching
		Run PIXY to check data used for matching.
	------------------------------------------------------ */

	if (args.length == 0) {
	    printHelp();
	    System.exit(0);
	}

	if (args[0].equals("-test")) {
	    if (args.length == 1) {
		System.out.println("java misao.pixy.Pixy -test <command>");
		System.out.println("command:");
		System.out.println("detect    Check raw detected stars.");
		System.out.println("matching  Check data used for matching.");
	    } else {
		testPixy(args[1]);
	    }
	} else if (args[0].equals("-diff")) {
//	    Difference diff = analyzeTwoLogFiles(args);
//	    diff.operate();

	    analyzeDiff(args);
	} else if (args[0].equals("-batch")) {
	    analyzeBatchFile(args[1]);
	} else if (args[0].equals("-identify")) {
	    analyzeIdentifier(args);
	} else if (args[0].equals("-report")) {
	    outputReport(args);
	} else if (args[0].equals("-survey")) {
	    analyzeNDChecker(args);
	} else if (args[0].equals("-update")) {
	    analyzeUpdater(args);
	} else {
	    PixyInputData input_data = null;
	    PixyOutputData output_data = null;

	    if (args[0].equals("-arg")) {
		input_data = analyzeArguments(args);
		output_data = runPixy(input_data);
	    } else if (args[0].equals("-console")) {
		input_data = analyzeQandA();
		output_data = runPixy(input_data);
	    } else if (args[0].equals("-review")) {
		output_data = analyzePxfFile(args);
	    } else {
		printHelp();
		System.exit(0);
	    }

	    MainFrame main_frame = new MainFrame(output_data);
	    main_frame.show();
	}
    }

    private static void printHelp ( ) {
	System.out.println("Options:");
	System.out.println("-arg <arguments>                        Run PIXY with parameters in arguments.");
	System.out.println("-batch <batch file>                     Run PIXY with parameters in batch file.");
	System.out.println("-console                                Run PIXY setting parameters in Q&A style.");
//	System.out.println("-diff <pxf file1> <pxf file2>           Output difference between two images.");
	System.out.println("-diff <image1> <image2> [<output_file>] Output difference between two images.");
	System.out.println("-identify <pxf file> [<check file>]     Identification with known variable stars.");
	System.out.println("-report <pxf file> [<report file>]      Output a photometric report.");
	System.out.println("-review <pxf file> [<image>]            Review PIXY with a PXF file.");
	System.out.println("-survey <directory> [<output file>]     List up candidates of new objects.");
	System.out.println("-test detect                            Run PIXY to check raw detected stars.");
	System.out.println("-test matching                          Run PIXY to check data used for matching.");
	System.out.println("-update [-remove] <pxf file>            Update magnitude database.");
    }

    private static PixyOutputData runPixy ( PixyInputData input_data ) throws Exception {
	RetryManager retry_manager = new RetryManager(input_data);

	while (input_data != null) {
	    try {
		if (retry_manager.getRetryCount() > 0)
		    System.out.println("Retrying...");

		PixyKernel kernel = new PixyKernel(input_data);
		kernel.outputToConsole(true);
		PixyOutputData output_data = (PixyOutputData)kernel.operate();
		return output_data;
	    } catch (Exception e) {
		input_data = retry_manager.getNextInputData();
	    }
	}

	return null;
    }

    private static void testPixy ( String arg ) throws Exception {
	PixyInputData input_data = analyzeQandA();

	PixyKernel kernel = new PixyKernel(input_data);
	kernel.outputToConsole(true);

	if (arg.equals("detect")) {
	    kernel.testDetect();
	    PixyOutputDataDetect output_data = (PixyOutputDataDetect)kernel.operate();
	    DetectFrame frame = new DetectFrame(output_data);
	    frame.show();
	} else if (arg.equals("matching")) {
	    kernel.testMatching();
	    PixyOutputDataMatching output_data = (PixyOutputDataMatching)kernel.operate();
	    MatchingFrame frame = new MatchingFrame(output_data);
	    frame.show();
	}
    }

    // java misao.pixy.Pixy -review <pxf file> [<image>]
    // args[1]: pxf file
    // args[2]: image file
    private static PixyOutputData analyzePxfFile ( String[] args ) throws Exception {
	File file = new File(args[1]);
	FileInputStream fis = new FileInputStream(file);
	DataInputStream dis = new DataInputStream(new BufferedInputStream(fis));
	PxfFile pxf = new PxfFile();
	pxf.read(dis);
	dis.close();

	String image_url = pxf.image_info.image_url;
	if (args.length > 2)
	    image_url = args[2];

	PixyOutputData output_data = new PixyOutputData();
	try {
	    URL url = new URL(new URLString(image_url).toString());
	    output_data.image = new ImagePixy(new ImageFile(url).getGrayImage());
	} catch ( FileNotFoundException e ) {
	    int p = image_url.lastIndexOf('/');
	    int q = image_url.lastIndexOf('\\');
	    if (q > p)
		p = q;
	    if (p >= 0)
		image_url = image_url.substring(p+1);
	    URL url = new URL(new URLString(image_url).toString());
	    output_data.image = new ImagePixy(new ImageFile(url).getGrayImage());
	}
	output_data.image_info = pxf.image_info;
	output_data.pair_list = pxf.pair_list;
	output_data.map_function = new MapFunction(1.0, 360.0 - pxf.image_info.pa, new Position());

	// reverse in case of a mirror image
	if (pxf.input_data.east_right)
	    output_data.image.mirrorHorizontal();

	// rescale the image in case of SBIG image (ST-4/6)
	if (pxf.input_data.sbig_ratio) {
	    int h = (int)((double)output_data.image.getHeight() * Astro.SBIG_RATIO + 0.5);
	    output_data.image = new ImagePixy(output_data.image.rescale(new Size(output_data.image.getWidth(), h)));
	}

	// (x,y) of all catalog data
	((PixyPairList)output_data.pair_list).setAllCatalogPosition(pxf.image_info.center_coor, (double)pxf.image_info.image_size.getWidth() / pxf.image_info.field_of_view);

	return output_data;
    }

/*
    // java misao.pixy.Pixy -diff <logfile1> <logfile2>
    // args[1]: log file 1
    // args[2]: log file 2
    private static Difference analyzeTwoLogFiles ( String[] args ) throws Exception {
	String[] s = new String[2];
	s[1] = args[1];
	PixyOutputData output_data1 = analyzeLogFile(s);
	s[1] = args[2];
	PixyOutputData output_data2 = analyzeLogFile(s);

	Difference diff = new Difference(output_data1, output_data2);

	return diff;
    }
*/

    // java misao.pixy.Pixy -diff <image1> <image2> [<output file>]
    private static void analyzeDiff ( String[] args ) throws Exception {
	String image1 = null;
	String image2 = null;
	String pxf_filename = null;
	String log_filename = null;
	boolean noPS = false;
	boolean auto = false;

	boolean check_image_flag = false;
	boolean check_gallery_flag = false;

	for (int i = 0 ; i < args.length ; i++) {
	    if (args[i].equals("-noPS")) {
		noPS = true;
	    } else if (args[i].equals("-auto")) {
		auto = true;
	    } else if (args[i].equals("-checkimage")) {
		check_image_flag = true;
	    } else if (args[i].equals("-checkgallery")) {
		check_gallery_flag = true;
	    } else if (args[i].charAt(0) == '-') {
	    } else if (image1 == null) {
		image1 = args[i];
	    } else if (image2 == null) {
		image2 = args[i];
	    } else if (pxf_filename == null) {
		pxf_filename = args[i];
	    } else if (log_filename == null) {
		log_filename = args[i];
	    }
	}

	DifferentialPixyInputData input_data = new DifferentialPixyInputData();
	input_data.image_url1 = image1;
	input_data.image_url2 = image2;
	if (pxf_filename != null)
	    input_data.pxf_filename = pxf_filename;
	if (log_filename != null)
	    input_data.output_filename = log_filename;

	PixyKernel kernel = new PixyKernel(null);
	kernel.outputToConsole(true);
	kernel.outputPSData(! noPS);
	DifferentialPixyOutputData output_data = kernel.difference(input_data);

	DifferentialCheckImage check_image = new DifferentialCheckImage(output_data);

	if (check_image_flag) {
	    String dir = ".";
	    if (pxf_filename != null) {
		dir = PxfFile.getFilenameBody(pxf_filename);
		File directory = new File(dir);
		if (! directory.isDirectory())
		    directory.mkdir();
	    }
	    check_image.createCheckImage(dir);
	}

	if (! (auto == true  &&  check_gallery_flag == false)) {
	    DifferentialGalleryFrame frame = new DifferentialGalleryFrame(output_data);
	    frame.show();

	    if (check_gallery_flag) {
		while (frame.getImage() == null) {
		    try {
			Thread.sleep(100);
		    } catch (InterruptedException e) {
			break;
		    }
		}

		ImageGray gallery = frame.getImage();
		File file = new File(PxfFile.getFilenameBody(pxf_filename) + ".fts");
		ImageFileFits fits = new ImageFileFits(gallery);
		fits.save(file);
	    }

	    System.out.println("Done.");

	    if (auto) {
		frame.dispose();
		frame.stop();
		System.exit(0);
	    }
	}
    }

    // java misao.pixy.Pixy -console
    // Q and A style
    private static PixyInputData analyzeQandA ( ) throws Exception {
	PixyInputData input_data = new PixyInputData();

	System.out.print("URL or filename of image: ");

	DataInputStream din = new DataInputStream(System.in);

	input_data.setParameter(new Parameter("image", din.readLine()));
	
	System.out.println("Exposure start date and time in UT: ");
	System.out.println("  ex. 1998 Jan. 31 01:23:45");
	System.out.println("  acceptable style: 1998 Jan 31 01:23:45");
	System.out.println("                    1998 1 31 1 23 45");
	System.out.print("Exposure start date and time in UT: ");
	String s = din.readLine().trim();
	if (s.length() > 0)
	    input_data.setParameter(new Parameter("exposure_start", s));

	System.out.print("Exposure in seconds: ");
	s = din.readLine().trim();
	if (s.length() > 0)
	    input_data.setParameter(new Parameter("exposure", s));

	System.out.println("List of accessible star catalogs: ");
	System.out.println("Code                    Catalog");
	CatalogIndex catalog_index = new CatalogIndex();
	Catalog catalog = catalog_index.getFirstAllskyCdromStarCatalog();
	while (catalog != null) {
	    String code = catalog.getCode() + "                        ";
	    code = code.substring(0, 24);
	    System.out.println(code + catalog.getName());
	    catalog = catalog_index.getNextAllskyCdromStarCatalog();
	}
	System.out.print("Input code: ");
	input_data.setParameter(new Parameter("catalog", din.readLine()));

	if (input_data.catalog.isCdrom()) {
	    System.out.print("Root directories of the catalog: ");
	    input_data.setParameter(new Parameter("catalog_path", din.readLine()));
	}

	System.out.println("Approximate position (J2000.0): ");
	System.out.println("  ex. R.A.= 12h 4m31.2s  Decl.= - 0o15'");
	System.out.println("  acceptable style: 12 4 31.2 -0 15 0");
	System.out.print("Approximate position (J2000.0): ");
	input_data.setParameter(new Parameter("center", din.readLine()));

	System.out.print("Position error in degree: ");
	s = din.readLine().trim();
	if (s.length() > 0)
	    input_data.setParameter(new Parameter("pos_error", s));

	System.out.print("Approximate horizontal field of view in degree: ");
	input_data.setParameter(new Parameter("field_of_view", din.readLine()));

	System.out.print("Limiting magnitude to be detected at least: ");
	s = din.readLine().trim().trim();
	if (s.length() > 0)
	    input_data.setParameter(new Parameter("limiting_mag", s));

	System.out.print("Magnitude code: ");
	input_data.setParameter(new Parameter("mag_code", din.readLine()));

	System.out.print("Mirror image? [y/N]: ");
	input_data.setParameter(new Parameter("mirror", din.readLine()));

	System.out.print("SBIG ratio (ST-4/6)? [y/N]: ");
	input_data.setParameter(new Parameter("sbig_ratio", din.readLine()));

	System.out.print("Reject cosmic ray? [y/N]: ");
	input_data.setParameter(new Parameter("reject_CR", din.readLine()));

	System.out.print("Print star data in detail (much memory are required)? [y/N]: ");
	input_data.setParameter(new Parameter("detail", din.readLine()));

	System.out.print("Filename to save midway progress (.log): ");
	input_data.setParameter(new Parameter("log", din.readLine()));

	System.out.print("Filename to save results (.pxf): ");
	input_data.setParameter(new Parameter("pxf", din.readLine()));

	return input_data;
    }

    // java misao.pixy.Pixy -arg <arguments>
    private static PixyInputData analyzeArguments ( String[] args ) throws Exception {
	PixyInputData input_data = new PixyInputData();

	for (int i = 0 ; i < args.length ; i++)
	    input_data.setParameter(new Parameter(args[i]));

	return input_data;
    }

    // java misao.pixy.Pixy -batch <batch file>
    private static void analyzeBatchFile ( String filename ) throws Exception {
	PixyInputData input_data = new PixyInputData();

	File file = new File(filename);
	FileInputStream fis = new FileInputStream(file);
	DataInputStream dis = new DataInputStream(new BufferedInputStream(fis));

	boolean all_done = true;

	String record = dis.readLine();
	while (record != null) {
	    record = record.trim();
	    if (record.length() == 0) {
		if (all_done == false)
		    operatePixyByBatchFile(input_data);

		all_done = true;
	    } else if (record.charAt(0) == ';'  ||  record.charAt(0) == '#') {
		// comment
	    } else {
		input_data.setParameter(new Parameter(record));
		all_done = false;
	    }
	    record = dis.readLine();
	}

	if (all_done == false)
	    operatePixyByBatchFile(input_data);

	dis.close();
    }

    // run PIXY for one image in batch file
    private static void operatePixyByBatchFile ( PixyInputData input_data ) {
	RetryManager retry_manager = new RetryManager(input_data);
	String image = input_data.image_url_string;

	while (input_data != null) {
	    try {
		if (retry_manager.getRetryCount() == 0)
		    System.out.println("Examining " + image);
		else
		    System.out.println("Retrying " + image);

		PixyKernel kernel = new PixyKernel(input_data);
		kernel.outputToConsole(false);
		kernel.setExceptionThrowableAtNoCdrom(true);
		PixyOutputData output_data = (PixyOutputData)kernel.operate();

		input_data = null;
	    } catch (Exception e) {
		input_data = retry_manager.getNextInputData();
		if (input_data == null)
		    System.err.println("ERROR: " + image);
	    }
	}
    }

    // java misao.pixy.Pixy -report <pxf file>
    private static void outputReport ( String[] args ) throws Exception {
	File file = new File(args[1]);
	FileInputStream fis = new FileInputStream(file);
	DataInputStream dis = new DataInputStream(new BufferedInputStream(fis));
	PxfFile pxf = new PxfFile();
	pxf.read(dis);
	dis.close();

	String report_file = null;
	if (args.length > 2)
	    report_file = args[2];

	PhotometricReporter reporter = new PhotometricReporter((PixyPairList)pxf.pair_list, pxf.image_info);
	reporter.outputToConsole(true);
	reporter.outputToReportFile(report_file);

	reporter.operate();
    }

    // java misao.pixy.Pixy -identify <pxf file>
    private static void analyzeIdentifier ( String[] args ) throws Exception {
	String file1 = null;
	String file2 = null;
	boolean batch_mode = false;
	boolean no_NR = false;

	for (int i = 0 ; i < args.length ; i++) {
	    if (args[i].equals("-batch")) {
		i++;
		file1 = args[i];
		batch_mode = true;
	    } else if (args[i].equals("-noNR")) {
		no_NR = true;
	    } else if (args[i].charAt(0) == '-') {
	    } else if (file1 == null) {
		file1 = args[i];
	    } else if (file2 == null) {
		file2 = args[i];
	    }
	}

	if (batch_mode) {
	    File file = new File(file1);
	    FileInputStream fis = new FileInputStream(file);
	    DataInputStream dis = new DataInputStream(new BufferedInputStream(fis));

	    String s = dis.readLine();
	    while (s != null) {
		StringTokenizer st = new StringTokenizer(s);
		file1 = st.nextToken();
		file2 = null;
		if (st.hasMoreTokens())
		    file2 = st.nextToken();

		System.out.println("Identifying " + file1);

		try {
		    analyzeIdentifierOne(file1, file2, no_NR);
		} catch (Exception e) {
		    System.err.println("ERROR: " + file1);
		}

		s = dis.readLine();
	    }

	    dis.close();
	} else {
	    analyzeIdentifierOne(file1, file2, no_NR);
	}
    }

    private static void analyzeIdentifierOne ( String pxf_file, String check_file, boolean no_NR ) throws Exception {
	File file = new File(pxf_file);
	FileInputStream fis = new FileInputStream(file);
	DataInputStream dis = new DataInputStream(new BufferedInputStream(fis));
	PxfFile pxf = new PxfFile();
	pxf.read(dis);
	dis.close();

	Identifier identifier = new Identifier((PixyPairList)pxf.pair_list, pxf.image_info);
	identifier.outputToConsole(true);
	identifier.outputToCheckFile(check_file);
	if (no_NR)
	    identifier.setPrintAllPair();

	pxf.pair_list = identifier.operate();

	// save PXF file
	pxf.pair_list.outputAllData(true);
	PrintManager pxf_manager = new PrintManager();
	pxf_manager.outputToConsole(false);
	pxf_manager.openFile(pxf_file);
	pxf_manager.printObject(pxf);
	pxf_manager.close();
    }

    // java misao.pixy.Pixy -survey <directory> [<output file>]
    private static void analyzeNDChecker ( String[] args ) throws Exception {
	File directory = null;
	String output_file = null;
	String pxf_path = null;

	for (int i = 0 ; i < args.length ; i++) {
	    if (args[i].equals("-pxf")) {
		i++;
		pxf_path = args[i];
	    } else if (args[i].charAt(0) == '-') {
	    } else if (directory == null) {
		directory = new File(args[i]);
	    } else if (output_file == null) {
		output_file = args[i];
	    }
	}

	NDChecker checker = new NDChecker(directory);
	checker.outputMidwayProgress(true);
	checker.outputToConsole(true);
	checker.outputToFile(output_file);

	if (pxf_path != null) {
	    StringTokenizer st = new StringTokenizer(pxf_path);
	    String[] paths = new String[st.countTokens()];
	    int i = 0;
	    while (st.hasMoreTokens())
		paths[i++] = st.nextToken();
	    checker.setPxfPath(paths);
	}

	checker.operate();
    }

    // java misao.pixy.Pixy -update [-remove] <pxf file>
    private static void analyzeUpdater ( String[] args ) throws Exception {
	String filename = null;
	String image_id = null;
	boolean batch_mode = false;
	boolean remove_flag = false;

	for (int i = 0 ; i < args.length ; i++) {
	    if (args[i].equals("-batch")) {
		i++;
		filename = args[i];
		batch_mode = true;
	    } else if (args[i].equals("-image")) {
		i++;
		image_id = args[i];
	    } else if (args[i].equals("-remove")) {
		remove_flag = true;
	    } else if (args[i].charAt(0) == '-') {
	    } else if (filename == null) {
		filename = args[i];
	    }
	}

	Updater.createDirectory();

	if (batch_mode) {
	    File file = new File(filename);
	    FileInputStream fis = new FileInputStream(file);
	    DataInputStream dis = new DataInputStream(new BufferedInputStream(fis));

	    String s = dis.readLine();
	    while (s != null) {
		StringTokenizer st = new StringTokenizer(s);
		filename = st.nextToken();
		image_id = "";
		if (st.hasMoreTokens())
		    image_id = st.nextToken();

		System.out.println("Updating " + filename);

		try {
		    analyzeUpdaterOne(filename, image_id, remove_flag);
		} catch (Exception e) {
		    System.err.println("ERROR: " + filename);
		}

		s = dis.readLine();
	    }

	    dis.close();
	} else {
	    analyzeUpdaterOne(filename, image_id, remove_flag);
	}
    }

    private static void analyzeUpdaterOne ( String pxf_file, String image_id, boolean remove_flag ) throws Exception {
	File file = new File(pxf_file);
	FileInputStream fis = new FileInputStream(file);
	DataInputStream dis = new DataInputStream(new BufferedInputStream(fis));
	PxfFile pxf = new PxfFile();
	pxf.read(dis);
	dis.close();

	Updater updater = new Updater(pxf);
	updater.setImageID(image_id);
	if (remove_flag == true)
	    updater.setToDelete();

	updater.operate();
    }
}
