/*
 * Orbit.java
 *
 * Copyright (C) 1998-2000 Seiichi Yoshida
 *
 */

package misao.util;

import java.lang.Math;
import misao.util.*;

public class Orbit {
    public JulianDay T;
    public double q;
    public double e;
    public double peri;
    public double node;
    public double i;
    public double M;

    static final double PRE = 0.0000001;
    static final double PRE_DATE = 0.00001;

    public Orbit ( ) {
	T = new JulianDay();
	q = 1.0;
	e = 1.0;
	peri = 0.0;
	node = 0.0;
	i = 0.0;
	M = 0.0;
    }

    public void convertAsteroidToComet ( ) {
	double a = q;
	double P = Math.sqrt(a*a*a);
	double n = Astro.GAUSS_K / Astro.RAD / P;
	if (M >= 180)
	    T = new JulianDay(T.getJD() + (360.0 - M) / n);
	else
	    T = new JulianDay(T.getJD() - M / n);

	q = q * (1.0 - e);
    }

    public OrbitPosition calculatePosition ( JulianDay jd ) {
	double a1,a2,a3,a4,a5,a6,a7,a8;
	double px,py,pz,qx,qy,qz;
	double ax,ay,az,bx,by,bz;
	double a,b;
	double M,M1,E;
	double sinE,cosE,tanv,xsi,eta,zeta;
	double x,y,z,r;

	ax = ay = az = bx = by = bz = 0.0;

	double dt = (Astro.get_t2000(jd.getJD()) - Astro.get_t2000(T.getJD())) * 36525.0;

	a1 = Math.sin(peri*Astro.RAD) * Math.sin(node*Astro.RAD);
	a2 = Math.sin(peri*Astro.RAD) * Math.cos(node*Astro.RAD);
	a3 = Math.sin(peri*Astro.RAD) * Math.sin(i*Astro.RAD);
	a4 = Math.cos(peri*Astro.RAD) * Math.sin(node*Astro.RAD);
	a5 = Math.cos(peri*Astro.RAD) * Math.cos(node*Astro.RAD);
	a6 = Math.cos(peri*Astro.RAD) * Math.sin(i*Astro.RAD);
	a7 = a5 * Math.cos(i*Astro.RAD) - a1;
	a8 = a2 * Math.cos(i*Astro.RAD) + a4;

	sinE = Math.sin(Astro.EPSILON*Astro.RAD);
	cosE = Math.cos(Astro.EPSILON*Astro.RAD);
	
	px = a5 - a1 * Math.cos(i*Astro.RAD);
	py = a8 * cosE - a3 * sinE;
	pz = a3 * cosE + a8 * sinE;
	qx = - a2 - a4 * Math.cos(i*Astro.RAD);
	qy = a7 * cosE - a6 * sinE;
	qz = a6 * cosE + a7 * sinE;

	if(e < 1.0){
	    a = q / (1.0 - e);
	    b = a * Math.sqrt(1.0 - e * e);
	    ax = a * px;
	    ay = a * py;
	    az = a * pz;
	    bx = b * qx;
	    by = b * qy;
	    bz = b * qz;
	}else if(e > 1.0){
	    a = Math.abs(q / (1.0 - e));
	    b = a * Math.sqrt(e * e - 1.0);
	    ax = a * px;
	    ay = a * py;
	    az = a * pz;
	    bx = b * qx;
	    by = b * qy;
	    bz = b * qz;
	}else{
	    ax = q * px;
	    ay = q * py;
	    az = q * pz;
	    bx = 2.0 * q * qx;
	    by = 2.0 * q * qy;
	    bz = 2.0 * q * qz;
	}

	SolarXyz solar_XYZ = new SolarXyz(jd);

	OrbitPosition position = new OrbitPosition();

	xsi = eta = zeta = 0.0;

	double dt0 = dt;
	double delay, delay_old;
	delay = 0.0;
	do{
	    delay_old = delay;

	    dt = dt0 - delay;

	    if(e < 1.0){
		a = q / (1.0 - e);
		M = Astro.GAUSS_K / Math.pow(a,1.5) * dt;
		E = kepler_equation(M,M / (1.0 - e),e);
		sinE = Math.sin(E);
		cosE = Math.cos(E);

		r = a * (1.0 - e * cosE);
		x = ax * (cosE - e) + bx * sinE;
		y = ay * (cosE - e) + by * sinE;
		z = az * (cosE - e) + bz * sinE;
	    }else if(e > 1.0){
		a = Math.abs(q / (1.0 - e));
		M = Astro.GAUSS_K / Math.pow(a,1.5) * dt;
		E = Astro.GAUSS_K * dt / (Math.sqrt(2.0) * Math.pow(q,1.5));
		E = kepler_equation_h(M,E,e);
		sinE = Astro.sinh(E);
		cosE = Astro.cosh(E);

		r = a * (e * cosE - 1.0);
		x = ax * (e - cosE) + bx * sinE;
		y = ay * (e - cosE) + by * sinE;
		z = az * (e - cosE) + bz * sinE;
	    }else{
		int count = 0;

		M = Astro.GAUSS_K * dt / (Math.sqrt(2.0) * Math.pow(q,1.5));
		tanv = M;
		do{
		    M1 = tanv * tanv * tanv + 3.0 * tanv - 3.0 * M;
		    tanv -= M1 / (3.0 * tanv * tanv + 3.0);
		    count++;
		    if(count > 20){
			tanv *= (0.5 + Math.random());
			count = 0;
		    }
		}while(Math.abs(M1) > PRE  ||  count == 0);

		r = q * (1.0 + tanv * tanv);
		x = ax * (1.0 - tanv * tanv) + bx * tanv;
		y = ay * (1.0 - tanv * tanv) + by * tanv;
		z = az * (1.0 - tanv * tanv) + bz * tanv;
	    }

	    position.r = r;

	    xsi = x + solar_XYZ.getX();
	    eta = y + solar_XYZ.getY();
	    zeta = z + solar_XYZ.getZ();
	    position.delta = Math.sqrt(xsi * xsi + eta * eta + zeta * zeta);

	    delay = position.delta * Astro.LIGHT_DELAY / 3600.0 / 24.0;
	}while(Math.abs(delay - delay_old) > PRE_DATE);

	double coor_d  = Math.asin(zeta / position.delta) / Astro.RAD;
	double coor_a = Math.atan2(eta,xsi) / Astro.RAD;
	if(coor_a < 0.0)
	    coor_a += 360.0;
	position.coor = new Coor(coor_a, coor_d);
	double solar_r = Math.sqrt(solar_XYZ.getX() * solar_XYZ.getX() + solar_XYZ.getY() * solar_XYZ.getY() + solar_XYZ.getZ() * solar_XYZ.getZ());
	position.elong = Math.acos((solar_r * solar_r + position.delta * position.delta - position.r * position.r) / (2.0 * solar_r * position.delta)) / Astro.RAD;

	return position;
    }

    public static double kepler_equation(double M,double E,double e) {
	double M1;
	double E0;
	int count = 0;

	E0 = E;
	do{
	    if(count > 20){
		E0 *= (0.5 + Math.random());
		E = E0;
		count = 0;
	    }
	    M1 = E - e * Math.sin(E);
	    E += (M - M1) / (1.0 - e * Math.cos(E));
	    count++;
	}while(Math.abs(M - M1) > PRE);

	return E;
    }

    public static double kepler_equation_h(double M,double E,double e) {
	double M1;
	double E0;
	int count = 0;
 
	E0 = E;
	do{
	    if(count > 20){
		E0 *= (0.5 + Math.random());
		E = E0;
		count = 0;
	    }
	    M1 = e * Astro.sinh(E) - E;
	    E += (M - M1) / (e * Astro.cosh(E) - 1.0);
	    count++;
	}while(Math.abs(M - M1) > PRE);
 
	return E;
    }
}
